# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: datasource.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import kismet_pb2 as kismet__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x64\x61tasource.proto\x12\x10KismetDatasource\x1a\x0ckismet.proto\"\x1f\n\x0bSubChannels\x12\x10\n\x08\x63hannels\x18\x01 \x03(\t\"\x1d\n\nSubChanset\x12\x0f\n\x07\x63hannel\x18\x01 \x02(\t\"c\n\nSubChanhop\x12\x10\n\x08\x63hannels\x18\x01 \x03(\t\x12\x0c\n\x04rate\x18\x02 \x01(\x01\x12\x0f\n\x07shuffle\x18\x03 \x01(\x08\x12\x14\n\x0cshuffle_skip\x18\x04 \x01(\r\x12\x0e\n\x06offset\x18\x05 \x01(\r\"\xc8\x01\n\x06SubGps\x12\x0b\n\x03lat\x18\x01 \x02(\x01\x12\x0b\n\x03lon\x18\x02 \x02(\x01\x12\x0b\n\x03\x61lt\x18\x03 \x02(\x01\x12\r\n\x05speed\x18\x04 \x02(\x01\x12\x0f\n\x07heading\x18\x05 \x02(\x01\x12\x11\n\tprecision\x18\x06 \x02(\x01\x12\x0b\n\x03\x66ix\x18\x07 \x02(\r\x12\x10\n\x08time_sec\x18\x08 \x02(\x04\x12\x11\n\ttime_usec\x18\t \x02(\x04\x12\x0c\n\x04type\x18\n \x02(\t\x12\x0c\n\x04name\x18\x0b \x02(\t\x12\x16\n\x0ehigh_prec_time\x18\x0c \x01(\x01\"X\n\x0cSubInterface\x12\x11\n\tinterface\x18\x01 \x02(\t\x12\r\n\x05\x66lags\x18\x02 \x02(\t\x12\x10\n\x08hardware\x18\x03 \x01(\t\x12\x14\n\x0c\x63\x61pinterface\x18\x04 \x01(\t\"k\n\tSubPacket\x12\x10\n\x08time_sec\x18\x01 \x02(\x04\x12\x11\n\ttime_usec\x18\x02 \x02(\x04\x12\x0b\n\x03\x64lt\x18\x03 \x02(\r\x12\x0c\n\x04size\x18\x04 \x02(\x04\x12\x0c\n\x04\x64\x61ta\x18\x05 \x02(\x0c\x12\x10\n\x08\x63\x61p_size\x18\x06 \x01(\x04\"J\n\x07SubJson\x12\x10\n\x08time_sec\x18\x01 \x02(\x04\x12\x11\n\ttime_usec\x18\x02 \x02(\x04\x12\x0c\n\x04type\x18\x03 \x02(\t\x12\x0c\n\x04json\x18\x04 \x02(\t\"N\n\tSubBuffer\x12\x10\n\x08time_sec\x18\x01 \x02(\x04\x12\x11\n\ttime_usec\x18\x02 \x02(\x04\x12\x0c\n\x04type\x18\x03 \x02(\t\x12\x0e\n\x06\x62uffer\x18\x04 \x02(\x0c\"\x90\x01\n\tSubSignal\x12\x12\n\nsignal_dbm\x18\x01 \x01(\x01\x12\x11\n\tnoise_dbm\x18\x02 \x01(\x01\x12\x13\n\x0bsignal_rssi\x18\x03 \x01(\x01\x12\x12\n\nnoise_rssi\x18\x04 \x01(\x01\x12\x10\n\x08\x66req_khz\x18\x05 \x01(\x01\x12\x0f\n\x07\x63hannel\x18\x06 \x01(\t\x12\x10\n\x08\x64\x61tarate\x18\x07 \x01(\x01\"\x9f\x01\n\nSubSpecset\x12\x11\n\tstart_mhz\x18\x01 \x01(\x01\x12\x0f\n\x07\x65nd_mhz\x18\x02 \x01(\x01\x12\x1a\n\x12samples_per_bucket\x18\x03 \x01(\x01\x12\x17\n\x0f\x62ucket_width_hz\x18\x04 \x01(\x01\x12\x12\n\nenable_amp\x18\x05 \x01(\x08\x12\x0e\n\x06if_amp\x18\x06 \x01(\x04\x12\x14\n\x0c\x62\x61seband_amp\x18\x07 \x01(\x04\"}\n\x0bSubSpectrum\x12\x10\n\x08time_sec\x18\x01 \x01(\x04\x12\x11\n\ttime_usec\x18\x02 \x01(\x04\x12\x11\n\tstart_mhz\x18\x03 \x01(\x01\x12\x0f\n\x07\x65nd_mhz\x18\x04 \x01(\x01\x12\x17\n\x0f\x62ucket_width_hz\x18\x05 \x01(\x01\x12\x0c\n\x04\x64\x61ta\x18\x06 \x03(\x05\",\n\nSubSuccess\x12\x0f\n\x07success\x18\x01 \x02(\x08\x12\r\n\x05seqno\x18\x02 \x02(\r\"\x11\n\x0f\x43loseDatasource\"\x99\x01\n\tConfigure\x12-\n\x07\x63hannel\x18\x01 \x01(\x0b\x32\x1c.KismetDatasource.SubChanset\x12-\n\x07hopping\x18\x02 \x01(\x0b\x32\x1c.KismetDatasource.SubChanhop\x12.\n\x08spectrum\x18\x03 \x01(\x0b\x32\x1c.KismetDatasource.SubSpecset\"\xdf\x01\n\x0f\x43onfigureReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12-\n\x07\x63hannel\x18\x02 \x01(\x0b\x32\x1c.KismetDatasource.SubChanset\x12-\n\x07hopping\x18\x03 \x01(\x0b\x32\x1c.KismetDatasource.SubChanhop\x12.\n\x07message\x18\x04 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12\x0f\n\x07warning\x18\x05 \x01(\t\"\xed\x02\n\nDataReport\x12%\n\x03gps\x18\x01 \x01(\x0b\x32\x18.KismetDatasource.SubGps\x12.\n\x07message\x18\x02 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12+\n\x06packet\x18\x03 \x01(\x0b\x32\x1b.KismetDatasource.SubPacket\x12+\n\x06signal\x18\x04 \x01(\x0b\x32\x1b.KismetDatasource.SubSignal\x12/\n\x08spectrum\x18\x05 \x01(\x0b\x32\x1d.KismetDatasource.SubSpectrum\x12\x0f\n\x07warning\x18\x06 \x01(\t\x12\'\n\x04json\x18\x07 \x01(\x0b\x32\x19.KismetDatasource.SubJson\x12+\n\x06\x62uffer\x18\x08 \x01(\x0b\x32\x1b.KismetDatasource.SubBuffer\x12\x16\n\x0ehigh_prec_time\x18\t \x01(\x01\"l\n\x0b\x45rrorReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12.\n\x07message\x18\x02 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\"\x10\n\x0eListInterfaces\"\xa5\x01\n\x10InterfacesReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12.\n\x07message\x18\x02 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12\x32\n\ninterfaces\x18\x03 \x03(\x0b\x32\x1e.KismetDatasource.SubInterface\"A\n\tNewSource\x12\x12\n\ndefinition\x18\x01 \x02(\t\x12\x12\n\nsourcetype\x18\x02 \x02(\t\x12\x0c\n\x04uuid\x18\x03 \x02(\t\" \n\nOpenSource\x12\x12\n\ndefinition\x18\x01 \x02(\t\"\x8c\x03\n\x10OpenSourceReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12\x0b\n\x03\x64lt\x18\x02 \x01(\r\x12\x19\n\x11\x63\x61pture_interface\x18\x03 \x01(\t\x12/\n\x08\x63hannels\x18\x04 \x01(\x0b\x32\x1d.KismetDatasource.SubChannels\x12-\n\x07\x63hannel\x18\x05 \x01(\x0b\x32\x1c.KismetDatasource.SubChanset\x12\x30\n\nhop_config\x18\x06 \x01(\x0b\x32\x1c.KismetDatasource.SubChanhop\x12\x10\n\x08hardware\x18\x07 \x01(\t\x12.\n\x07message\x18\x08 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12.\n\x08spectrum\x18\t \x01(\x0b\x32\x1c.KismetDatasource.SubSpecset\x12\x0c\n\x04uuid\x18\n \x01(\t\x12\x0f\n\x07warning\x18\x0b \x01(\t\"!\n\x0bProbeSource\x12\x12\n\ndefinition\x18\x01 \x02(\t\"\x94\x02\n\x11ProbeSourceReport\x12-\n\x07success\x18\x01 \x02(\x0b\x32\x1c.KismetDatasource.SubSuccess\x12.\n\x07message\x18\x02 \x01(\x0b\x32\x1d.KismetExternal.MsgbusMessage\x12/\n\x08\x63hannels\x18\x03 \x01(\x0b\x32\x1d.KismetDatasource.SubChannels\x12-\n\x07\x63hannel\x18\x04 \x01(\x0b\x32\x1c.KismetDatasource.SubChanset\x12.\n\x08spectrum\x18\x05 \x01(\x0b\x32\x1c.KismetDatasource.SubSpecset\x12\x10\n\x08hardware\x18\x06 \x01(\t\" \n\rWarningReport\x12\x0f\n\x07warning\x18\x01 \x02(\tB\x02H\x03')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'datasource_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'H\003'
  _SUBCHANNELS._serialized_start=52
  _SUBCHANNELS._serialized_end=83
  _SUBCHANSET._serialized_start=85
  _SUBCHANSET._serialized_end=114
  _SUBCHANHOP._serialized_start=116
  _SUBCHANHOP._serialized_end=215
  _SUBGPS._serialized_start=218
  _SUBGPS._serialized_end=418
  _SUBINTERFACE._serialized_start=420
  _SUBINTERFACE._serialized_end=508
  _SUBPACKET._serialized_start=510
  _SUBPACKET._serialized_end=617
  _SUBJSON._serialized_start=619
  _SUBJSON._serialized_end=693
  _SUBBUFFER._serialized_start=695
  _SUBBUFFER._serialized_end=773
  _SUBSIGNAL._serialized_start=776
  _SUBSIGNAL._serialized_end=920
  _SUBSPECSET._serialized_start=923
  _SUBSPECSET._serialized_end=1082
  _SUBSPECTRUM._serialized_start=1084
  _SUBSPECTRUM._serialized_end=1209
  _SUBSUCCESS._serialized_start=1211
  _SUBSUCCESS._serialized_end=1255
  _CLOSEDATASOURCE._serialized_start=1257
  _CLOSEDATASOURCE._serialized_end=1274
  _CONFIGURE._serialized_start=1277
  _CONFIGURE._serialized_end=1430
  _CONFIGUREREPORT._serialized_start=1433
  _CONFIGUREREPORT._serialized_end=1656
  _DATAREPORT._serialized_start=1659
  _DATAREPORT._serialized_end=2024
  _ERRORREPORT._serialized_start=2026
  _ERRORREPORT._serialized_end=2134
  _LISTINTERFACES._serialized_start=2136
  _LISTINTERFACES._serialized_end=2152
  _INTERFACESREPORT._serialized_start=2155
  _INTERFACESREPORT._serialized_end=2320
  _NEWSOURCE._serialized_start=2322
  _NEWSOURCE._serialized_end=2387
  _OPENSOURCE._serialized_start=2389
  _OPENSOURCE._serialized_end=2421
  _OPENSOURCEREPORT._serialized_start=2424
  _OPENSOURCEREPORT._serialized_end=2820
  _PROBESOURCE._serialized_start=2822
  _PROBESOURCE._serialized_end=2855
  _PROBESOURCEREPORT._serialized_start=2858
  _PROBESOURCEREPORT._serialized_end=3134
  _WARNINGREPORT._serialized_start=3136
  _WARNINGREPORT._serialized_end=3168
# @@protoc_insertion_point(module_scope)
